# -*- coding: utf-8 -*-

from . import plugin
from .core import Sijax

__title__ = 'sijax'
__version__ = '0.2.0'
__author__ = 'Slavi Pantaleev'
__license__ = 'BSD'
__copyright__ = 'Copyright 2011 Slavi Pantaleev'

